@@ -2,7 +2,7 @@
  * Create a squashfs filesystem.  This is a highly compressed read only
  * filesystem.
  *
- * Copyright (c) 2012, 2013, 2014, 2021, 2022
+ * Copyright (c) 2012, 2013, 2014, 2021, 2022, 2025
  * Phillip Lougher <phillip@squashfs.org.uk>
  *
  * This program is free software; you can redistribute it and/or
@@ -55,6 +55,7 @@ static int need_nl = FALSE;
 static int rotate = 0;
 static long long cur_uncompressed = 0, estimated_uncompressed = 0;
 static int columns;
+static double inc, base;
 
 static pthread_t progress_thread;
 static pthread_mutex_t progress_mutex = PTHREAD_MUTEX_INITIALIZER;
@@ -102,6 +103,28 @@ void progress_bar_size(int count)
 }
 
 
+void progress_bar_metadata(int inodes)
+{
+	int extra = estimated_uncompressed / 19;
+
+	inc = (double) extra / inodes;
+	base = estimated_uncompressed;
+
+	pthread_cleanup_push((void *) pthread_mutex_unlock, &size_mutex);
+	pthread_mutex_lock(&size_mutex);
+	estimated_uncompressed += extra;
+	pthread_cleanup_pop(1);
+}
+
+
+void inc_meta_progress_bar()
+{
+	base += inc;
+	if(cur_uncompressed != ((long long) base))
+		cur_uncompressed = (long long) base;
+}
+
+
 static void progressbar(long long current, long long max, int columns)
 {
 	char rotate_list[] = { '|', '/', '-', '\\' };
@@ -182,8 +205,10 @@ void enable_progress_bar()
 {
 	pthread_cleanup_push((void *) pthread_mutex_unlock, &progress_mutex);
 	pthread_mutex_lock(&progress_mutex);
-	if(display_progress_bar)
+	if(display_progress_bar) {
 		progress_bar(cur_uncompressed, estimated_uncompressed, columns);
+		need_nl = TRUE;
+	}
 	temp_disabled = FALSE;
 	pthread_cleanup_pop(1);
 }
@@ -219,6 +244,19 @@ void set_progressbar_state(int state)
 }
 
 
+void progressbar_finish()
+{
+	pthread_cleanup_push((void *) pthread_mutex_unlock, &progress_mutex);
+	pthread_mutex_lock(&progress_mutex);
+	if(display_progress_bar && !temp_disabled) {
+		progress_bar(estimated_uncompressed, estimated_uncompressed, columns);
+		printf("\n");
+	}
+	display_progress_bar = FALSE;
+	pthread_cleanup_pop(1);
+}
+
+
 static void *progress_thrd(void *arg)
 {
 	struct timespec requested_time, remaining;
