@@ -2,7 +2,8 @@
  * Create a squashfs filesystem.  This is a highly compressed read only
  * filesystem.
  *
- * Copyright (c) 2008, 2009, 2010, 2012, 2014, 2019, 2021, 2022
+ * Copyright (c) 2008, 2009, 2010, 2012, 2013, 2014, 2019, 2021, 2022, 2023,
+ * 2024, 2025
  * Phillip Lougher <phillip@squashfs.org.uk>
  *
  * This program is free software; you can redistribute it and/or
@@ -36,26 +37,21 @@
 #include <dirent.h>
 #include <string.h>
 #include <stdlib.h>
-#include <sys/xattr.h>
 #include <regex.h>
 
 #include "squashfs_fs.h"
 #include "squashfs_swap.h"
 #include "mksquashfs.h"
 #include "xattr.h"
+#include "mksquashfs_xattr.h"
 #include "mksquashfs_error.h"
 #include "progressbar.h"
 #include "pseudo.h"
 #include "tar.h"
 #include "action.h"
 #include "merge_sort.h"
-
-#ifdef XATTR_NOFOLLOW /* Apple's xattrs */
-	#define llistxattr(path_, buf_, sz_) \
-		listxattr(path_, buf_, sz_, XATTR_NOFOLLOW)
-	#define lgetxattr(path_, name_, val_, sz_) \
-		getxattr(path_, name_, val_, sz_, 0, XATTR_NOFOLLOW)
-#endif
+#include "alloc.h"
+#include "virt_disk_pos.h"
 
 /* compressed xattr table */
 static char *xattr_table = NULL;
@@ -91,7 +87,6 @@ static int xattr_add_count = 0;
 
 /* file system globals from mksquashfs.c */
 extern int no_xattrs, noX;
-extern long long bytes;
 extern int fd;
 extern unsigned int xattr_bytes, total_xattr_bytes;
 extern regex_t *xattr_exclude_preg;
@@ -126,7 +121,7 @@ static int xattr_get_type(char *name)
 
 static void xattr_copy_prefix(struct xattr_list *xattr, int t, char *name)
 {
-	xattr->full_name = strdup(name);
+	xattr->full_name = STRDUP(name);
 	xattr->name = xattr->full_name + strlen(prefix_table[t].prefix);
 	xattr->size = strlen(xattr->name);
 }
@@ -142,163 +137,6 @@ int xattr_get_prefix(struct xattr_list *xattr, char *name)
 	return type;
 }
 
-	
-static int read_xattrs_from_system(struct dir_ent *dir_ent, char *filename,
-						struct xattr_list **xattrs)
-{
-	ssize_t size, vsize;
-	char *xattr_names, *p;
-	int i = 0;
-	struct xattr_list *xattr_list = NULL;
-	struct xattr_data *xattr_exc_list;
-	struct xattr_data *xattr_inc_list;
-
-	while(1) {
-		size = llistxattr(filename, NULL, 0);
-		if(size <= 0) {
-			if(size < 0 && errno != ENOTSUP) {
-				ERROR_START("llistxattr for %s failed in "
-					"read_attrs, because %s", filename,
-					strerror(errno));
-				ERROR_EXIT(".  Ignoring\n");
-			}
-			return 0;
-		}
-
-		xattr_names = malloc(size);
-		if(xattr_names == NULL)
-			MEM_ERROR();
-
-		size = llistxattr(filename, xattr_names, size);
-		if(size < 0) {
-			free(xattr_names);
-			if(errno == ERANGE)
-				/* xattr list grew?  Try again */
-				continue;
-			else {
-				ERROR_START("llistxattr for %s failed in "
-					"read_attrs, because %s", filename,
-					strerror(errno));
-				ERROR_EXIT(".  Ignoring\n");
-				return 0;
-			}
-		}
-
-		break;
-	}
-
-	xattr_exc_list = eval_xattr_exc_actions(root_dir, dir_ent);
-	xattr_inc_list = eval_xattr_inc_actions(root_dir, dir_ent);
-
-	for(p = xattr_names; p < xattr_names + size;) {
-		struct xattr_list *x;
-		int res;
-
-		res = match_xattr_exc_actions(xattr_exc_list, p);
-		if(res) {
-			p += strlen(p) + 1;
-			continue;
-		}
-
-		if(xattr_exclude_preg) {
-			res = regexec(xattr_exclude_preg, p, (size_t) 0, NULL, 0);
-			if(res == 0) {
-				p += strlen(p) + 1;
-				continue;
-			}
-		}
-
-		res = match_xattr_inc_actions(xattr_inc_list, p);
-		if(res) {
-			p += strlen(p) + 1;
-			continue;
-		}
-
-		if(xattr_include_preg) {
-			res = regexec(xattr_include_preg, p, (size_t) 0, NULL, 0);
-			if(res) {
-				p += strlen(p) + 1;
-				continue;
-			}
-		}
-
-		x = realloc(xattr_list, (i + 1) * sizeof(struct xattr_list));
-		if(x == NULL)
-			MEM_ERROR();
-		xattr_list = x;
-
-		xattr_list[i].type = xattr_get_prefix(&xattr_list[i], p);
-
-		if(xattr_list[i].type == -1) {
-			ERROR("Unrecognised xattr prefix %s\n", p);
-			p += strlen(p) + 1;
-			continue;
-		}
-
-		p += strlen(p) + 1;
-
-		while(1) {
-			vsize = lgetxattr(filename, xattr_list[i].full_name,
-								NULL, 0);
-			if(vsize < 0) {
-				ERROR_START("lgetxattr failed for %s in "
-					"read_attrs, because %s", filename,
-					strerror(errno));
-				ERROR_EXIT(".  Ignoring\n");
-				free(xattr_list[i].full_name);
-				goto failed;
-			}
-
-			xattr_list[i].value = malloc(vsize);
-			if(xattr_list[i].value == NULL)
-				MEM_ERROR();
-
-			vsize = lgetxattr(filename, xattr_list[i].full_name,
-						xattr_list[i].value, vsize);
-			if(vsize < 0) {
-				free(xattr_list[i].value);
-				if(errno == ERANGE)
-					/* xattr grew?  Try again */
-					continue;
-				else {
-					ERROR_START("lgetxattr failed for %s "
-						"in read_attrs, because %s",
-						filename, strerror(errno));
-					ERROR_EXIT(".  Ignoring\n");
-					free(xattr_list[i].full_name);
-					goto failed;
-				}
-			}
-			
-			break;
-		}
-
-		xattr_list[i].vsize = vsize;
-
-		TRACE("read_xattrs_from_system: filename %s, xattr name %s,"
-			" vsize %d\n", filename, xattr_list[i].full_name,
-			xattr_list[i].vsize);
-		i++;
-	}
-
-	free(xattr_names);
-
-	if(i > 0)
-		*xattrs = xattr_list;
-	else
-		free(xattr_list);
-	return i;
-
-failed:
-	while(--i >= 0) {
-		free(xattr_list[i].full_name);
-		free(xattr_list[i].value);
-	}
-	free(xattr_list);
-	free(xattr_names);
-	return 0;
-}
-
 
 static int get_xattr_size(struct xattr_list *xattr)
 {
@@ -325,10 +163,8 @@ static void *get_xattr_space(unsigned int req_size, long long *disk)
 	while(cache_bytes >= SQUASHFS_METADATA_SIZE) {
 		if((xattr_size - xattr_bytes) <
 				((SQUASHFS_METADATA_SIZE << 1)) + 2) {
-			xattr_table = realloc(xattr_table, xattr_size +
+			xattr_table = REALLOC(xattr_table, xattr_size +
 				(SQUASHFS_METADATA_SIZE << 1) + 2);
-			if(xattr_table == NULL)
-				MEM_ERROR();
 			xattr_size += (SQUASHFS_METADATA_SIZE << 1) + 2;
 		}
 
@@ -349,10 +185,8 @@ static void *get_xattr_space(unsigned int req_size, long long *disk)
 	data_space = cache_size - cache_bytes;
 	if(data_space < req_size) {
 			int realloc_size = req_size - data_space;
-			data_cache = realloc(data_cache, cache_size +
+			data_cache = REALLOC(data_cache, cache_size +
 				realloc_size);
-			if(data_cache == NULL)
-				MEM_ERROR();
 			cache_size += realloc_size;
 	}
 
@@ -403,9 +237,7 @@ static struct dupl_id *check_id_dupl(struct xattr_list *xattr_list, int xattrs)
 
 	if(entry == NULL) {
 		/* no duplicate exists */
-		entry = malloc(sizeof(*entry));
-		if(entry == NULL)
-			MEM_ERROR();
+		entry = MALLOC(sizeof(*entry));
 		entry->xattrs = xattrs;
 		entry->xattr_list = xattr_list;
 		entry->xattr_id = SQUASHFS_INVALID_XATTR;
@@ -465,10 +297,8 @@ static int get_xattr_id(int xattrs, struct xattr_list *xattr_list,
 	int i, size = 0;
 	struct squashfs_xattr_id *xattr_id;
 
-	xattr_id_table = realloc(xattr_id_table, (xattr_ids + 1) *
+	xattr_id_table = REALLOC(xattr_id_table, (xattr_ids + 1) *
 		sizeof(struct squashfs_xattr_id));
-	if(xattr_id_table == NULL)
-		MEM_ERROR();
 
 	/* get total uncompressed size of xattr data, needed for stat */
 	for(i = 0; i < xattrs; i++)
@@ -496,7 +326,7 @@ long long write_xattrs()
 	unsigned short c_byte;
 	int i, avail_bytes;
 	char *datap = data_cache;
-	long long start_bytes = bytes;
+	long long start_bytes;
 	struct squashfs_xattr_table header = {};
 
 	if(xattr_ids == 0)
@@ -508,10 +338,8 @@ long long write_xattrs()
 	while(cache_bytes) {
 		if((xattr_size - xattr_bytes) <
 				((SQUASHFS_METADATA_SIZE << 1)) + 2) {
-			xattr_table = realloc(xattr_table, xattr_size +
+			xattr_table = REALLOC(xattr_table, xattr_size +
 				(SQUASHFS_METADATA_SIZE << 1) + 2);
-			if(xattr_table == NULL)
-				MEM_ERROR();
 			xattr_size += (SQUASHFS_METADATA_SIZE << 1) + 2;
 		}
 
@@ -529,8 +357,8 @@ long long write_xattrs()
 	/*
 	 * Write compressed xattr table to file system
 	 */
-	write_destination(fd, bytes, xattr_bytes, xattr_table);
-        bytes += xattr_bytes;
+	start_bytes = get_and_inc_dpos(xattr_bytes);
+	write_destination(fd, start_bytes, xattr_bytes, xattr_table);
 
 	/*
 	 * Swap if necessary the xattr id table
@@ -547,7 +375,7 @@ long long write_xattrs()
 }
 
 
-void free_xattr_list(int xattrs, struct xattr_list *xattr_list)
+static void free_xattr_list(int xattrs, struct xattr_list *xattr_list)
 {
 	int i;
 
@@ -560,7 +388,7 @@ void free_xattr_list(int xattrs, struct xattr_list *xattr_list)
 }
 
 
-int generate_xattrs(int xattrs, struct xattr_list *xattr_list)
+static int generate_xattrs(int xattrs, struct xattr_list *xattr_list)
 {
 	int total_size, i;
 	int xattr_value_max;
@@ -750,7 +578,6 @@ int read_xattrs(void *d, int type)
 	}
 
 	while(l1 || l2 || l3) {
-		struct xattr_list *x;
 		struct xattr_add *entry;
 
 		if(l1 && l2 && l3) {
@@ -815,18 +642,11 @@ int read_xattrs(void *d, int type)
 				 type != SQUASHFS_DIR_TYPE))
 			continue;
 
-		x = realloc(xattr_list, (i + 1) * sizeof(struct xattr_list));
-		if(x == NULL)
-			MEM_ERROR();
-		xattr_list = x;
-
+		xattr_list = REALLOC(xattr_list, (i + 1) * sizeof(struct xattr_list));
 		xattr_list[i].type = entry->type;
 		xattr_copy_prefix(&xattr_list[i], entry->type, entry->name);
 
-		xattr_list[i].value = malloc(entry->vsize);
-		if(xattr_list[i].value == NULL)
-			MEM_ERROR();
-
+		xattr_list[i].value = MALLOC(entry->vsize);
 		memcpy(xattr_list[i].value, entry->value, entry->vsize);
 		xattr_list[i].vsize = entry->vsize;
 
@@ -926,10 +746,7 @@ void save_xattrs()
 	 * Note we have to save the contents of the data cache because future
 	 * operations will delete the current contents
 	 */
-	sdata_cache = malloc(cache_bytes);
-	if(sdata_cache == NULL)
-		MEM_ERROR();
-
+	sdata_cache = MALLOC(cache_bytes);
 	memcpy(sdata_cache, data_cache, cache_bytes);
 	scache_bytes = cache_bytes;
 
@@ -959,10 +776,7 @@ void restore_xattrs()
 regex_t *xattr_regex(char *pattern, char *option)
 {
 	int error;
-	regex_t *preg = malloc(sizeof(regex_t));
-
-	if(preg == NULL)
-		MEM_ERROR();
+	regex_t *preg = MALLOC(sizeof(regex_t));
 
 	error = regcomp(preg, pattern, REG_EXTENDED|REG_NOSUB);
 
@@ -997,7 +811,7 @@ char *base64_decode(char *source, int size, int *bytes)
 	/* Calculate number of bytes the base64 encoding represents */
 	count = size * 3 / 4;
 
-	dest = malloc(count);
+	dest = MALLOC(count);
 
 	for(dest_ptr = (unsigned char *) dest; size; size --, source_ptr ++) {
 		int value = *source_ptr;
@@ -1051,7 +865,7 @@ failed:
 }
 
 
-char *hex_decode(char *source, int size, int *bytes)
+static char *hex_decode(char *source, int size, int *bytes)
 {
 	char *dest;
 	unsigned char *dest_ptr, *source_ptr = (unsigned char *) source;
@@ -1060,9 +874,7 @@ char *hex_decode(char *source, int size, int *bytes)
 	if(size % 2 != 0)
 		return NULL;
 
-	dest = malloc(size >> 2);
-	if(dest == NULL)
-		MEM_ERROR();
+	dest = MALLOC(size >> 2);
 
 	for(dest_ptr = (unsigned char *) dest ; size; size --) {
 		int digit = *source_ptr ++;
@@ -1092,7 +904,7 @@ failed:
 }
 
 
-int decode_octal(unsigned char *ptr)
+static int decode_octal(unsigned char *ptr)
 {
 	int i, output = 0;
 
@@ -1109,7 +921,7 @@ int decode_octal(unsigned char *ptr)
 }
 
 
-char *text_decode(char *source, int *bytes)
+static char *text_decode(char *source, int *bytes)
 {
 	unsigned char *dest, *dest_ptr, *ptr = (unsigned char *) source;
 	int size = 0;
@@ -1123,10 +935,7 @@ char *text_decode(char *source, int *bytes)
 		}
 	}
 
-	dest = malloc(size);
-	if(dest == NULL)
-		MEM_ERROR();
-
+	dest = MALLOC(size);
 	*bytes = size;
 
 	for(ptr = (unsigned char *) source, dest_ptr = dest; size; size --) {
@@ -1178,11 +987,8 @@ struct xattr_add *xattr_parse(char *str, char *pre, char *option)
 		goto failed;
 	}
 
-	entry = malloc(sizeof(struct xattr_add));
-	if(entry == NULL)
-		MEM_ERROR();
-
-	entry->name = strndup(str, value++ - str);
+	entry = MALLOC(sizeof(struct xattr_add));
+	entry->name = STRNDUP(str, value++ - str);
 	entry->type = xattr_get_type(entry->name);
 
 	if(entry->type == -1) {
@@ -1215,13 +1021,13 @@ struct xattr_add *xattr_parse(char *str, char *pre, char *option)
 		}
 
 		entry->value = base64_decode(value, strlen(value), &size);
-		entry->vsize = size;
-
 		if(entry->value == NULL) {
 			ERROR("%sinvalid argument %s in %s option, because "
 				"invalid base64 value\n", pre, str, option);
 			goto failed2;
 		}
+
+		entry->vsize = size;
 		break;
 
 	case PREFIX_HEX_0X:
@@ -1235,13 +1041,13 @@ struct xattr_add *xattr_parse(char *str, char *pre, char *option)
 		}
 
 		entry->value = hex_decode(value, strlen(value), &size);
-		entry->vsize = size;
-
 		if(entry->value == NULL) {
 			ERROR("%sinvalid argument %s in %s option, because "
 				"invalid hexidecimal value\n", pre, str, option);
 			goto failed2;
 		}
+
+		entry->vsize = size;
 		break;
 
 	case PREFIX_TEXT_0T:
@@ -1255,13 +1061,13 @@ struct xattr_add *xattr_parse(char *str, char *pre, char *option)
 		}
 
 		entry->value = text_decode(value, &size);
-		entry->vsize = size;
-
 		if(entry->value == NULL) {
 			ERROR("%sinvalid argument %s in %s option, because "
 				"invalid text value\n", pre, str, option);
 			goto failed2;
 		}
+
+		entry->vsize = size;
 		break;
 
 	case PREFIX_BINARY_0B:
@@ -1277,11 +1083,7 @@ struct xattr_add *xattr_parse(char *str, char *pre, char *option)
 		/* fall through */
 	default:
 		entry->vsize = strlen(value);
-		entry->value = malloc(entry->vsize);
-
-		if(entry->value == NULL)
-			MEM_ERROR();
-
+		entry->value = MALLOC(entry->vsize);
 		memcpy(entry->value, value, entry->vsize);
 	}
 
